@@ -1,84 +1,75 @@
-/*
-Copyright (c) 2012 Sony Pictures Imageworks Inc., et al.
-and
-All Rights Reserved.
-
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are
-met:
-* Redistributions of source code must retain the above copyright
-  notice, this list of conditions and the following disclaimer.
-* Redistributions in binary form must reproduce the above copyright
-  notice, this list of conditions and the following disclaimer in the
-  documentation and/or other materials provided with the distribution.
-* Neither the name of Sony Pictures Imageworks nor the names of its
-  contributors may be used to endorse or promote products derived from
-  this software without specific prior written permission.
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-*/
+// Copyright Contributors to the Open Shading Language project.
+// SPDX-License-Identifier: BSD-3-Clause
+// https://github.com/imageworks/OpenShadingLanguage
 
 
 // Extend Imath matrix classes to include 2x2, as analogous as possible
 // to Imath::Matrix33 and Imath::Matrix44.
-// 
+//
 // The original Imath classes bear the "new BSD" license (same as
 // ours above) and this copyright:
 // Copyright (c) 2002, Industrial Light & Magic, a division of
 // Lucas Digital Ltd. LLC.  All rights reserved.
+// SPDX-License-Identifier: BSD-3-Clause
 
 
 #pragma once
 
 
-#include <OpenEXR/ImathMatrix.h>
+#include <OSL/oslconfig.h>
+
+#include <OSL/Imathx/ImathMatrix.h>
+#include <limits>
+
 
 namespace Imathx {   // "extended" Imath
 
 enum Uninitialized {UNINITIALIZED};
 
 
+
 template <class T> class Matrix22
 {
   public:
+    typedef T		        BaseType;
+    typedef Imath::Vec2<T>	BaseVecType;
 
     //-------------------
     // Access to elements
     //-------------------
 
     T           x[2][2];
-
+#if 0 // deprecate operator [] usage to force use of data member vs. 2 function calls that loose type info along the way
+    OSL_HOSTDEVICE
     T *         operator [] (int i);
-    const T *   operator [] (int i) const;
 
+    OSL_HOSTDEVICE
+    const T *   operator [] (int i) const;
+#endif
 
     //-------------
     // Constructors
     //-------------
 
+    OSL_HOSTDEVICE
     Matrix22 (Uninitialized) {}
 
+    OSL_HOSTDEVICE
     Matrix22 ();
                                 // 1 0
                                 // 0 1
 
+    OSL_HOSTDEVICE
     Matrix22 (T a);
                                 // a a
                                 // a a
 
+    OSL_HOSTDEVICE
     Matrix22 (const T a[2][2]);
                                 // a[0][0] a[0][1]
                                 // a[1][0] a[1][1]
 
+    OSL_HOSTDEVICE
     Matrix22 (T a, T b, T c, T d);
                                 // a b
                                 // c d
@@ -88,26 +79,42 @@ template <class T> class Matrix22
     // Copy constructor and assignment
     //--------------------------------
 
+    OSL_HOSTDEVICE
     Matrix22 (const Matrix22 &v);
-    template <class S> explicit Matrix22 (const Matrix22<S> &v);
 
+    template <class S> OSL_HOSTDEVICE explicit
+    Matrix22 (const Matrix22<S> &v);
+
+    OSL_HOSTDEVICE
     const Matrix22 &    operator = (const Matrix22 &v);
+
+    OSL_HOSTDEVICE
     const Matrix22 &    operator = (T a);
 
 
+    //------------
+    // Destructor
+    //------------
+
+    ~Matrix22 () = default;
+
     //----------------------
     // Compatibility with Sb
     //----------------------
     
+    OSL_HOSTDEVICE
     T *                 getValue ();
+
+    OSL_HOSTDEVICE
     const T *           getValue () const;
 
-    template <class S>
+    template <class S> OSL_HOSTDEVICE
     void                getValue (Matrix22<S> &v) const;
-    template <class S>
+
+    template <class S> OSL_HOSTDEVICE
     Matrix22 &          setValue (const Matrix22<S> &v);
 
-    template <class S>
+    template <class S> OSL_HOSTDEVICE
     Matrix22 &          setTheMatrix (const Matrix22<S> &v);
 
 
@@ -115,6 +122,7 @@ template <class T> class Matrix22
     // Identity
     //---------
 
+    OSL_HOSTDEVICE
     void                makeIdentity();
 
 
@@ -122,7 +130,10 @@ template <class T> class Matrix22
     // Equality
     //---------
 
+    OSL_HOSTDEVICE
     bool                operator == (const Matrix22 &v) const;
+
+    OSL_HOSTDEVICE
     bool                operator != (const Matrix22 &v) const;
 
     //-----------------------------------------------------------------------
@@ -143,7 +154,10 @@ template <class T> class Matrix22
     //      abs (this[i] - v[i][j]) <= e * abs (this[i][j])
     //-----------------------------------------------------------------------
 
+    OSL_HOSTDEVICE
     bool                equalWithAbsError (const Matrix22<T> &v, T e) const;
+
+    OSL_HOSTDEVICE
     bool                equalWithRelError (const Matrix22<T> &v, T e) const;
 
 
@@ -151,8 +165,13 @@ template <class T> class Matrix22
     // Component-wise addition
     //------------------------
 
+    OSL_HOSTDEVICE
     const Matrix22 &    operator += (const Matrix22 &v);
+
+    OSL_HOSTDEVICE
     const Matrix22 &    operator += (T a);
+
+    OSL_HOSTDEVICE
     Matrix22            operator + (const Matrix22 &v) const;
 
 
@@ -160,8 +179,13 @@ template <class T> class Matrix22
     // Component-wise subtraction
     //---------------------------
 
+    OSL_HOSTDEVICE
     const Matrix22 &    operator -= (const Matrix22 &v);
+
+    OSL_HOSTDEVICE
     const Matrix22 &    operator -= (T a);
+
+    OSL_HOSTDEVICE
     Matrix22            operator - (const Matrix22 &v) const;
 
 
@@ -169,7 +193,10 @@ template <class T> class Matrix22
     // Component-wise multiplication by -1
     //------------------------------------
 
+    OSL_HOSTDEVICE
     Matrix22            operator - () const;
+
+    OSL_HOSTDEVICE
     const Matrix22 &    negate ();
 
 
@@ -177,7 +204,10 @@ template <class T> class Matrix22
     // Component-wise multiplication
     //------------------------------
 
+    OSL_HOSTDEVICE
     const Matrix22 &    operator *= (T a);
+
+    OSL_HOSTDEVICE
     Matrix22            operator * (T a) const;
 
 
@@ -185,7 +215,10 @@ template <class T> class Matrix22
     // Matrix-times-matrix multiplication
     //-----------------------------------
 
+    OSL_HOSTDEVICE
     const Matrix22 &    operator *= (const Matrix22 &v);
+
+    OSL_HOSTDEVICE
     Matrix22            operator * (const Matrix22 &v) const;
 
 
@@ -201,7 +234,7 @@ template <class T> class Matrix22
     // submatrix, ignoring the rest of matrix m.
     //-----------------------------------------------------------------
 
-    template <class S>
+    template <class S> OSL_HOSTDEVICE
     void                multMatrix(const Imath::Vec2<S> &src, Imath::Vec2<S> &dst) const;
 
 
@@ -210,7 +243,10 @@ template <class T> class Matrix22
     // Component-wise division
     //------------------------
 
+    OSL_HOSTDEVICE
     const Matrix22 &    operator /= (T a);
+
+    OSL_HOSTDEVICE
     Matrix22            operator / (T a) const;
 
 
@@ -218,7 +254,10 @@ template <class T> class Matrix22
     // Transposed matrix
     //------------------
 
+    OSL_HOSTDEVICE
     const Matrix22 &    transpose ();
+
+    OSL_HOSTDEVICE
     Matrix22            transposed () const;
 
 
@@ -231,16 +270,21 @@ template <class T> class Matrix22
     // 
     //------------------------------------------------------------
 
+    OSL_HOSTDEVICE
     const Matrix22 &    invert (bool singExc = false);
 
-    Matrix22<T>         inverse (bool singExc = false) const;
+    OSL_HOSTDEVICE    
+    Matrix22<T>         inverse () const noexcept;
 
+    OSL_HOSTDEVICE
+    Matrix22<T>         inverse (bool singExc) const;
+    
 
     //-----------------------------------------
     // Set matrix to rotation by r (in radians)
     //-----------------------------------------
 
-    template <class S>
+    template <class S> OSL_HOSTDEVICE
     const Matrix22 &    setRotation (S r);
 
 
@@ -248,7 +292,7 @@ template <class T> class Matrix22
     // Rotate the given matrix by r
     //-----------------------------
 
-    template <class S>
+    template <class S> OSL_HOSTDEVICE
     const Matrix22 &    rotate (S r);
 
 
@@ -256,6 +300,7 @@ template <class T> class Matrix22
     // Set matrix to scale by given uniform factor
     //--------------------------------------------
 
+    OSL_HOSTDEVICE
     const Matrix22 &    setScale (T s);
 
 
@@ -263,7 +308,7 @@ template <class T> class Matrix22
     // Set matrix to scale by given vector
     //------------------------------------
 
-    template <class S>
+    template <class S> OSL_HOSTDEVICE
     const Matrix22 &    setScale (const Imath::Vec2<S> &s);
 
 
@@ -271,7 +316,7 @@ template <class T> class Matrix22
     // Scale the matrix by s
     //----------------------
 
-    template <class S>
+    template <class S> OSL_HOSTDEVICE
     const Matrix22 &    scale (const Imath::Vec2<S> &s);
 
 
@@ -279,9 +324,16 @@ template <class T> class Matrix22
     // Limitations of type T (see also class limits<T>)
     //-------------------------------------------------
 
+    OSL_HOSTDEVICE
     static T            baseTypeMin()           {return Imath::limits<T>::min();}
+
+    OSL_HOSTDEVICE
     static T            baseTypeMax()           {return Imath::limits<T>::max();}
+
+    OSL_HOSTDEVICE
     static T            baseTypeSmallest()      {return Imath::limits<T>::smallest();}
+
+    OSL_HOSTDEVICE
     static T            baseTypeEpsilon()       {return Imath::limits<T>::epsilon();}
 
   private:
@@ -313,7 +365,7 @@ std::ostream &  operator << (std::ostream & s, const Matrix22<T> &m);
 // Vector-times-matrix multiplication operators
 //---------------------------------------------
 
-template <class S, class T>
+template <class S, class T> OSL_HOSTDEVICE
 Imath::Vec2<S> operator * (const Matrix22<T> &m, const Imath::Vec2<S> &v)
 {
     Imath::Vec2<S> tmp;
@@ -334,22 +386,23 @@ typedef Matrix22 <double> M22d;
 //---------------------------
 // Implementation of Matrix22
 //---------------------------
-
-template <class T>
+#if 0 // deprecated
+template <class T> OSL_HOSTDEVICE
 inline T *
 Matrix22<T>::operator [] (int i)
 {
     return x[i];
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 inline const T *
 Matrix22<T>::operator [] (int i) const
 {
     return x[i];
 }
+#endif
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 inline
 Matrix22<T>::Matrix22 ()
 {
@@ -359,7 +412,7 @@ Matrix22<T>::Matrix22 ()
     x[1][1] = 1;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 inline
 Matrix22<T>::Matrix22 (T a)
 {
@@ -369,14 +422,20 @@ Matrix22<T>::Matrix22 (T a)
     x[1][1] = a;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 inline
-Matrix22<T>::Matrix22 (const T a[2][2]) 
+Matrix22<T>::Matrix22 (const T a[2][2])
+: x{{a[0][0],
+     a[0][1]},
+    {a[1][0],
+     a[1][1]}}
 {
-    memcpy (x, a, sizeof (x));
+    // function calls and aliasing issues can inhibit vectorization
+    // versus straight assignment of data members
+    //memcpy (x, a, sizeof (x));
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 inline
 Matrix22<T>::Matrix22 (T a, T b, T c, T d)
 {
@@ -386,15 +445,21 @@ Matrix22<T>::Matrix22 (T a, T b, T c, T d)
     x[1][1] = d;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 inline
 Matrix22<T>::Matrix22 (const Matrix22 &v)
+: x{{v.x[0][0],
+     v.x[0][1]},
+    {v.x[1][0],
+     v.x[1][1]}}
 {
-    memcpy (x, v.x, sizeof (x));
+    // function calls and aliasing issues can inhibit vectorization
+    // versus straight assignment of data members
+    //memcpy (x, v.x, sizeof (x));
 }
 
 template <class T>
-template <class S>
+template <class S> OSL_HOSTDEVICE
 inline
 Matrix22<T>::Matrix22 (const Matrix22<S> &v)
 {
@@ -404,15 +469,19 @@ Matrix22<T>::Matrix22 (const Matrix22<S> &v)
     x[1][1] = T (v.x[1][1]);
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 inline const Matrix22<T> &
 Matrix22<T>::operator = (const Matrix22 &v)
 {
-    memcpy (x, v.x, sizeof (x));
+    //memcpy (x, v.x, sizeof (x));
+    x[0][0] = v.x[0][0];
+    x[0][1] = v.x[0][1];
+    x[1][0] = v.x[1][0];
+    x[1][1] = v.x[1][1];
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 inline const Matrix22<T> &
 Matrix22<T>::operator = (T a)
 {
@@ -423,14 +492,14 @@ Matrix22<T>::operator = (T a)
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 inline T *
 Matrix22<T>::getValue ()
 {
     return (T *) &x[0][0];
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 inline const T *
 Matrix22<T>::getValue () const
 {
@@ -438,15 +507,19 @@ Matrix22<T>::getValue () const
 }
 
 template <class T>
-template <class S>
+template <class S> OSL_HOSTDEVICE
 inline void
 Matrix22<T>::getValue (Matrix22<S> &v) const
 {
+    // function calls and aliasing issues can inhibit vectorization
+    // versus straight assignment of data members
+#if 0
     if (isSameType<S,T>::value)
     {
         memcpy (v.x, x, sizeof (x));
     }
     else
+#endif
     {
         v.x[0][0] = x[0][0];
         v.x[0][1] = x[0][1];
@@ -456,15 +529,19 @@ Matrix22<T>::getValue (Matrix22<S> &v) const
 }
 
 template <class T>
-template <class S>
+template <class S> OSL_HOSTDEVICE
 inline Matrix22<T> &
 Matrix22<T>::setValue (const Matrix22<S> &v)
 {
+    // function calls and aliasing issues can inhibit vectorization
+    // versus straight assignment of data members
+#if 0
     if (isSameType<S,T>::value)
     {
         memcpy (x, v.x, sizeof (x));
     }
     else
+#endif
     {
         x[0][0] = v.x[0][0];
         x[0][1] = v.x[0][1];
@@ -476,15 +553,19 @@ Matrix22<T>::setValue (const Matrix22<S> &v)
 }
 
 template <class T>
-template <class S>
+template <class S> OSL_HOSTDEVICE
 inline Matrix22<T> &
 Matrix22<T>::setTheMatrix (const Matrix22<S> &v)
 {
+    // function calls and aliasing issues can inhibit vectorization
+    // versus straight assignment of data members
+#if 0
     if (isSameType<S,T>::value)
     {
         memcpy (x, v.x, sizeof (x));
     }
     else
+#endif
     {
         x[0][0] = v.x[0][0];
         x[0][1] = v.x[0][1];
@@ -495,7 +576,7 @@ Matrix22<T>::setTheMatrix (const Matrix22<S> &v)
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 inline void
 Matrix22<T>::makeIdentity()
 {
@@ -505,7 +586,7 @@ Matrix22<T>::makeIdentity()
     x[1][1] = 1;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 bool
 Matrix22<T>::operator == (const Matrix22 &v) const
 {
@@ -515,7 +596,7 @@ Matrix22<T>::operator == (const Matrix22 &v) const
            x[1][1] == v.x[1][1];
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 bool
 Matrix22<T>::operator != (const Matrix22 &v) const
 {
@@ -525,31 +606,47 @@ Matrix22<T>::operator != (const Matrix22 &v) const
            x[1][1] != v.x[1][1];
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 bool
 Matrix22<T>::equalWithAbsError (const Matrix22<T> &m, T e) const
 {
+    // Simplify control flow by unrolling loop completely
+#if 0
     for (int i = 0; i < 2; i++)
         for (int j = 0; j < 2; j++)
-            if (!Imath::equalWithAbsError ((*this)[i][j], m[i][j], e))
+            if (!Imath::equalWithAbsError ((*this).x[i][j], m.x[i][j], e))
                 return false;
 
     return true;
+#else
+    return  Imath::equalWithAbsError ((*this).x[0][0], m.x[0][0], e) &&
+            Imath::equalWithAbsError ((*this).x[0][1], m.x[0][1], e) &&
+            Imath::equalWithAbsError ((*this).x[1][0], m.x[1][0], e) &&
+            Imath::equalWithAbsError ((*this).x[1][1], m.x[1][1], e);
+#endif
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 bool
 Matrix22<T>::equalWithRelError (const Matrix22<T> &m, T e) const
 {
+    // Simplify control flow by unrolling loop completely
+#if 0
     for (int i = 0; i < 2; i++)
         for (int j = 0; j < 2; j++)
-            if (!Imath::equalWithRelError ((*this)[i][j], m[i][j], e))
+            if (!Imath::equalWithRelError ((*this).x[i][j], m.x[i][j], e))
                 return false;
 
     return true;
+#else
+    return  Imath::equalWithRelError ((*this).x[0][0], m.x[0][0], e) &&
+            Imath::equalWithRelError ((*this).x[0][1], m.x[0][1], e) &&
+            Imath::equalWithRelError ((*this).x[1][0], m.x[1][0], e) &&
+            Imath::equalWithRelError ((*this).x[1][1], m.x[1][1], e);
+#endif
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::operator += (const Matrix22<T> &v)
 {
@@ -561,7 +658,7 @@ Matrix22<T>::operator += (const Matrix22<T> &v)
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::operator += (T a)
 {
@@ -573,7 +670,7 @@ Matrix22<T>::operator += (T a)
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 Matrix22<T>
 Matrix22<T>::operator + (const Matrix22<T> &v) const
 {
@@ -583,7 +680,7 @@ Matrix22<T>::operator + (const Matrix22<T> &v) const
                      x[1][1] + v.x[1][1]);
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::operator -= (const Matrix22<T> &v)
 {
@@ -595,7 +692,7 @@ Matrix22<T>::operator -= (const Matrix22<T> &v)
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::operator -= (T a)
 {
@@ -607,7 +704,7 @@ Matrix22<T>::operator -= (T a)
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 Matrix22<T>
 Matrix22<T>::operator - (const Matrix22<T> &v) const
 {
@@ -617,7 +714,7 @@ Matrix22<T>::operator - (const Matrix22<T> &v) const
                      x[1][1] - v.x[1][1]);
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 Matrix22<T>
 Matrix22<T>::operator - () const
 {
@@ -627,7 +724,7 @@ Matrix22<T>::operator - () const
                      -x[1][1]);
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::negate ()
 {
@@ -639,7 +736,7 @@ Matrix22<T>::negate ()
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::operator *= (T a)
 {
@@ -651,7 +748,7 @@ Matrix22<T>::operator *= (T a)
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 Matrix22<T>
 Matrix22<T>::operator * (T a) const
 {
@@ -661,57 +758,85 @@ Matrix22<T>::operator * (T a) const
                      x[1][1] * a);
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 inline Matrix22<T>
 operator * (T a, const Matrix22<T> &v)
 {
     return v * a;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::operator *= (const Matrix22<T> &v)
 {
+    // Manually unroll loops and avoid writing to
+    // the data members twice by using the uninitialized
+    // version of the constructor
+#if 0
     Matrix22 tmp (T (0));
 
     for (int i = 0; i < 2; i++)
         for (int j = 0; j < 2; j++)
             for (int k = 0; k < 2; k++)
                 tmp.x[i][j] += x[i][k] * v.x[k][j];
+#else
+    Matrix22 tmp (Imathx::UNINITIALIZED);
+
+    tmp.x[0][0] = x[0][0] * v.x[0][0] + x[0][1] * v.x[1][0];
+    tmp.x[0][1] = x[0][0] * v.x[0][1] + x[0][1] * v.x[1][1];
+
+    tmp.x[1][0] = x[1][0] * v.x[0][0] + x[1][1] * v.x[1][0];
+    tmp.x[1][1] = x[1][0] * v.x[0][1] + x[1][1] * v.x[1][1];
+#endif
 
     *this = tmp;
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 Matrix22<T>
 Matrix22<T>::operator * (const Matrix22<T> &v) const
 {
-    Matrix22 tmp (T (0));
 
+    // Manually unroll loops and avoid writing to
+    // the data members twice by using the uninitialized
+    // version of the constructor
+#if 0
+    Matrix22 tmp (T (0));
     for (int i = 0; i < 2; i++)
         for (int j = 0; j < 2; j++)
             for (int k = 0; k < 2; k++)
                 tmp.x[i][j] += x[i][k] * v.x[k][j];
+#else
+    Matrix22 tmp (Imathx::UNINITIALIZED);
+    
+    tmp.x[0][0] = x[0][0] * v.x[0][0] + x[0][1] * v.x[1][0];
+    tmp.x[0][1] = x[0][0] * v.x[0][1] + x[0][1] * v.x[1][1];
+    
+    tmp.x[1][0] = x[1][0] * v.x[0][0] + x[1][1] * v.x[1][0];
+    tmp.x[1][1] = x[1][0] * v.x[0][1] + x[1][1] * v.x[1][1];
+    
+    
+#endif
+    
 
     return tmp;
 }
 
 template <class T>
-template <class S>
+template <class S> OSL_HOSTDEVICE
 void
 Matrix22<T>::multMatrix(const Imath::Vec2<S> &src, Imath::Vec2<S> &dst) const
 {
-    S a, b;
-
-    a = src[0] * x[0][0] + src[1] * x[1][0];
-    b = src[0] * x[0][1] + src[1] * x[1][1];
+    // Non-aliasing by avoiding [int] operator on Vec2
+    S a = src.x * x[0][0] + src.y * x[1][0];
+    S b = src.x * x[0][1] + src.y * x[1][1];
 
     dst.x = a;
     dst.y = b;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::operator /= (T a)
 {
@@ -723,7 +848,7 @@ Matrix22<T>::operator /= (T a)
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 Matrix22<T>
 Matrix22<T>::operator / (T a) const
 {
@@ -733,7 +858,7 @@ Matrix22<T>::operator / (T a) const
                      x[1][1] / a);
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::transpose ()
 {
@@ -745,7 +870,7 @@ Matrix22<T>::transpose ()
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 Matrix22<T>
 Matrix22<T>::transposed () const
 {
@@ -756,7 +881,7 @@ Matrix22<T>::transposed () const
 }
 
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::invert (bool singExc)
 {
@@ -764,7 +889,46 @@ Matrix22<T>::invert (bool singExc)
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
+Matrix22<T>
+Matrix22<T>::inverse () const noexcept
+{
+    Matrix22 s ( x[1][1],  -x[0][1],
+                -x[1][0],   x[0][0]);
+    T r = x[0][0] * x[1][1] - x[1][0] * x[0][1];  // determinant
+
+    T abs_r = Imath::abs (r);
+    
+    int may_have_divided_by_zero = 0;
+    if (OSL_UNLIKELY(abs_r < T(1)))
+    {
+        float mr = abs_r / Imath::limits<T>::smallest();
+
+        if ((mr <= Imath::abs (s.x[0][0])) ||
+            (mr <= Imath::abs (s.x[0][1])) ||
+            (mr <= Imath::abs (s.x[1][0])) ||
+            (mr <= Imath::abs (s.x[1][1])))
+        {
+            may_have_divided_by_zero = 1;
+        }
+    }
+
+    if (OSL_UNLIKELY(may_have_divided_by_zero == 1))
+    {
+        //s = Matrix22();
+        s.x[0][0] = T(1);
+        s.x[0][1] = T(0);
+        s.x[1][0] = T(0);
+        s.x[1][1] = T(1);
+    } else {
+        s /= r;
+    }
+
+    return s;
+}
+
+
+template <class T> OSL_HOSTDEVICE
 Matrix22<T>
 Matrix22<T>::inverse (bool singExc) const
 {
@@ -772,27 +936,31 @@ Matrix22<T>::inverse (bool singExc) const
                 -x[1][0],   x[0][0]);
     T r = x[0][0] * x[1][1] - x[1][0] * x[0][1];  // determinant
 
-    if (Imath::abs (r) >= 1)
+    if (IMATH_INTERNAL_NAMESPACE::abs (r) >= 1)
     {
         s /= r;
     }
     else
     {
-        T mr = Imath::abs (r) / Imath::limits<T>::smallest();
+        T mr = IMATH_INTERNAL_NAMESPACE::abs (r) / Imath::limits<T>::smallest();
 
+        OSL_INTEL_PRAGMA(unroll)
         for (int i = 0; i < 2; ++i)
         {
+            OSL_INTEL_PRAGMA(unroll)
             for (int j = 0; j < 2; ++j)
             {
-                if (mr > Imath::abs (s[i][j]))
+                if (mr > IMATH_INTERNAL_NAMESPACE::abs (s.x[i][j]))
                 {
-                    s[i][j] /= r;
+                    s.x[i][j] /= r;
                 }
                 else
                 {
+#ifndef __CUDA_ARCH__
                     if (singExc)
                         throw Imath::SingMatrixExc ("Cannot invert "
                                              "singular matrix.");
+#endif
                     return Matrix22();
                 }
             }
@@ -803,12 +971,14 @@ Matrix22<T>::inverse (bool singExc) const
 }
 
 template <class T>
-template <class S>
+template <class S> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::setRotation (S r)
 {
     S cos_r, sin_r;
 
+    // TODO: consider calling std::cos/sin to allow compiler to use its own
+    // optimized version.
     cos_r = Imath::Math<T>::cos (r);
     sin_r = Imath::Math<T>::sin (r);
 
@@ -822,7 +992,7 @@ Matrix22<T>::setRotation (S r)
 }
 
 template <class T>
-template <class S>
+template <class S> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::rotate (S r)
 {
@@ -830,39 +1000,51 @@ Matrix22<T>::rotate (S r)
     return *this;
 }
 
-template <class T>
+template <class T> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::setScale (T s)
 {
-    memset (x, 0, sizeof (x));
+    // function calls and aliasing issues can inhibit vectorization
+    // versus straight assignment of data members
+//    memset (x, 0, sizeof (x));
+    x[0][1] = static_cast<T>(0);
     x[0][0] = s;
+    x[1][0] = static_cast<T>(0);
     x[1][1] = s;
 
     return *this;
 }
 
 template <class T>
-template <class S>
+template <class S> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::setScale (const Imath::Vec2<S> &s)
 {
-    memset (x, 0, sizeof (x));
-    x[0][0] = s[0];
-    x[1][1] = s[1];
+    // function calls and aliasing issues can inhibit vectorization
+    // versus straight assignment of data members
+    //memset (x, 0, sizeof (x));
+    //x[0][0] = s[0];
+    //x[1][1] = s[1];
+    x[0][0] = s.x;
+    x[0][1] = static_cast<T>(0);
+
+    x[1][0] = static_cast<T>(0);
+    x[1][1] = s.y;
 
     return *this;
 }
 
 template <class T>
-template <class S>
+template <class S> OSL_HOSTDEVICE
 const Matrix22<T> &
 Matrix22<T>::scale (const Imath::Vec2<S> &s)
 {
-    x[0][0] *= s[0];
-    x[0][1] *= s[0];
+    // avoid aliasing of Vec2::operator[](int)
+    x[0][0] *= s.x;
+    x[0][1] *= s.x;
 
-    x[1][0] *= s[1];
-    x[1][1] *= s[1];
+    x[1][0] *= s.y;
+    x[1][1] *= s.y;
 
     return *this;
 }
@@ -891,11 +1073,11 @@ operator << (std::ostream &s, const Matrix22<T> &m)
         width = s.precision() + 8;
     }
 
-    s << "(" << std::setw (width) << m[0][0] <<
-         " " << std::setw (width) << m[0][1] << "\n" <<
+    s << "(" << std::setw (width) << m.x[0][0] <<
+         " " << std::setw (width) << m.x[0][1] << "\n" <<
 
-         " " << std::setw (width) << m[1][0] <<
-         " " << std::setw (width) << m[1][1] << ")\n";
+         " " << std::setw (width) << m.x[1][0] <<
+         " " << std::setw (width) << m.x[1][1] << ")\n";
 
     s.flags (oldFlags);
     return s;
