@@ -1,30 +1,8 @@
-/*
-Copyright (c) 2009-2014 Sony Pictures Imageworks Inc., et al.
-All Rights Reserved.
+// Copyright Contributors to the Open Shading Language project.
+// SPDX-License-Identifier: BSD-3-Clause
+// https://github.com/imageworks/OpenShadingLanguage
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are
-met:
-* Redistributions of source code must retain the above copyright
-  notice, this list of conditions and the following disclaimer.
-* Redistributions in binary form must reproduce the above copyright
-  notice, this list of conditions and the following disclaimer in the
-  documentation and/or other materials provided with the distribution.
-* Neither the name of Sony Pictures Imageworks nor the names of its
-  contributors may be used to endorse or promote products derived from
-  this software without specific prior written permission.
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-*/
+// clang-format off
 
 #pragma once
 
@@ -44,13 +22,17 @@ struct ShaderGlobals;
 
 /// Opaque pointer to whatever the renderer uses to represent a
 /// (potentially motion-blurred) coordinate transformation.
-typedef const void * TransformationPtr;
+typedef const void* TransformationPtr;
 
 
 // Callbacks for closure creation
-typedef void (*PrepareClosureFunc)(RendererServices *, int id, void *data);
-typedef void (*SetupClosureFunc)(RendererServices *, int id, void *data);
+typedef void (*PrepareClosureFunc)(RendererServices*, int id, void* data);
+typedef void (*SetupClosureFunc)(RendererServices*, int id, void* data);
 
+// Turn off warnings about unused params for this file, since we have lots
+// of declarations with stub function bodies.
+OSL_PRAGMA_WARNING_PUSH
+OSL_GCC_PRAGMA(GCC diagnostic ignored "-Wunused-parameter")
 
 
 
@@ -62,7 +44,7 @@ public:
     typedef TextureSystem::Perthread TexturePerthread;
 
 
-    RendererServices (TextureSystem *texsys=NULL);
+    RendererServices(TextureSystem* texsys = NULL);
     virtual ~RendererServices () { }
 
     /// Given the name of a 'feature', return whether this RendererServices
@@ -147,7 +129,7 @@ public:
     /// the name of one of the coordinate systems).  A renderer is free
     /// to not implement this, in which case the default implementation
     /// is simply to make appropriate calls to get_matrix and
-    /// get_inverse_matrix.  The existance of this method is to allow
+    /// get_inverse_matrix.  The existence of this method is to allow
     /// some renderers to provide transformations that cannot be
     /// expressed by a 4x4 matrix.
     ///
@@ -200,17 +182,14 @@ public:
 
     /// Given the name of a texture, return an opaque handle that can be
     /// used with texture calls to avoid the name lookups.
-    virtual TextureHandle * get_texture_handle (ustring filename);
+    virtual TextureHandle * get_texture_handle (ustring filename,
+                                                ShadingContext *context);
 
     /// Return true if the texture handle (previously returned by
     /// get_texture_handle()) is a valid texture that can be subsequently
     /// read or sampled.
     virtual bool good (TextureHandle *texture_handle);
 
-    /// Return a thread-specific opaque pointer to the texture system.
-    /// Knowing the context may help this go faster.
-    virtual TexturePerthread * get_texture_perthread (ShadingContext *context=NULL);
-
     /// Filtered 2D texture lookup for a single point.
     ///
     /// s,t are the texture coordinates; dsdx, dtdx, dsdy, and dtdy are
@@ -244,15 +223,6 @@ public:
                           float dsdy, float dtdy, int nchannels,
                           float *result, float *dresultds, float *dresultdt,
                           ustring *errormessage);
-    // DEPRECATED(1.8) version, with no errormessage parameter. This will
-    // eventually disappear.
-    OSL_DEPRECATED ("Deprecated since 1.8, use the version with an errormessage parameter.")
-    virtual bool texture (ustring filename, TextureHandle *texture_handle,
-                          TexturePerthread *texture_thread_info,
-                          TextureOpt &options, ShaderGlobals *sg,
-                          float s, float t, float dsdx, float dtdx,
-                          float dsdy, float dtdy, int nchannels,
-                          float *result, float *dresultds, float *dresultdt);
 
     /// Filtered 3D texture lookup for a single point.
     ///
@@ -287,16 +257,6 @@ public:
                             float *result, float *dresultds,
                             float *dresultdt, float *dresultdr,
                             ustring *errormessage);
-    // DEPRECATED(1.8) version, with no errormessage parameter. This will
-    // eventually disappear.
-    OSL_DEPRECATED ("Deprecated since 1.8, use the version with an errormessage parameter.")
-    virtual bool texture3d (ustring filename, TextureHandle *texture_handle,
-                            TexturePerthread *texture_thread_info,
-                            TextureOpt &options, ShaderGlobals *sg,
-                            const Vec3 &P, const Vec3 &dPdx, const Vec3 &dPdy,
-                            const Vec3 &dPdz, int nchannels,
-                            float *result, float *dresultds,
-                            float *dresultdt, float *dresultdr);
 
     /// Filtered environment lookup for a single point.
     ///
@@ -327,15 +287,6 @@ public:
                               int nchannels, float *result,
                               float *dresultds, float *dresultdt,
                               ustring *errormessage);
-    // DEPRECATED(1.8) version, with no errormessage parameter. This will
-    // eventually disappear.
-    OSL_DEPRECATED ("Deprecated since 1.8, use the version with an errormessage parameter.")
-    virtual bool environment (ustring filename, TextureHandle *texture_handle,
-                              TexturePerthread *texture_thread_info,
-                              TextureOpt &options, ShaderGlobals *sg,
-                              const Vec3 &R, const Vec3 &dRdx, const Vec3 &dRdy,
-                              int nchannels, float *result,
-                              float *dresultds, float *dresultdt);
 
     /// Get information about the given texture.  Return true if found
     /// and the data has been put in *data.  Return false if the texture
@@ -350,14 +301,20 @@ public:
     /// can) use that extra information to perform a less expensive texture
     /// lookup.
     ///
-    /// Note to renderers: if sg is NULL, that means get_texture_info is
-    /// being called speculatively by the runtime optimizer, and it doesn't
-    /// know which object the shader will be run on.
-    virtual bool get_texture_info (ShaderGlobals *sg, ustring filename,
+    /// If the errormessage parameter is NULL, this method is expected to
+    /// handle the errors fully, including forwarding them to the renderer
+    /// or shading system. If errormessage is non-NULL, any resulting error
+    /// messages (in case of failure, when the function returns false) will
+    /// be stored there, leaving it up to the caller/shader to handle the
+    /// error.
+    virtual bool get_texture_info (ustring filename,
                                    TextureHandle *texture_handle,
+                                   TexturePerthread *texture_thread_info,
+                                   ShadingContext *shading_context,
                                    int subimage,
                                    ustring dataname, TypeDesc datatype,
-                                   void *data);
+                                   void *data,
+                                   ustring *errormessage);
 
 
     /// Lookup nearest points in a point cloud. It will search for
@@ -430,6 +387,20 @@ public:
         return 0;
     }
 
+    virtual uint64_t register_global (const std::string& var_name,
+                                      uint64_t           value)
+    {
+        return 0;
+    }
+
+    virtual bool fetch_global (const std::string& var_name,
+                               uint64_t*          value)
+    {
+        return false;
+    }
+
+
+
     /// Options we use for noise calls.
     struct NoiseOpt {
         int anisotropic;
@@ -446,4 +417,5 @@ protected:
 };
 
 
+OSL_PRAGMA_WARNING_POP
 OSL_NAMESPACE_EXIT
